verbose = False

QGIS_is_live = True


def log(message, note="", icon="🌐"):
    print(
        "{} {}{}".format(
            icon,
            (
                (f"{message:.<40}" if len(message) < 38 else f"{message}\n   {40*'.'}")
                if note
                else message
            ),
            note,
        )
    )


def log_error(message, note=""):
    log(message, note, icon="❗️")


def log_warning(message, note=""):
    log(message, note, icon="❓")



if not QGIS_is_live:
    from application import BLUE_GEO_QGIS_APPLICATION
    from project import project
    from seed import seed
    from QGIS import QGIS


class BLUE_GEO_QGIS_APPLICATION_UKRAINE_TIMEMAP(BLUE_GEO_QGIS_APPLICATION):
    def __init__(self):
        super().__init__("ukraine_timemap", "🇺🇦")

    def help(self):
        self.log(
            'ukraine_timemap.update("dryrun,~upload")',
            f"update {project.name}.",
        )

    def update(self, options: str = ""):
        self.log(f"ukraine_timemap.update({options}): {project.name}")

        seed(
            [
                "ukraine_timemap",
                "ingest",
                f"~copy_template,{options}",
                project.name,
            ]
        )


ukraine_timemap = BLUE_GEO_QGIS_APPLICATION_UKRAINE_TIMEMAP()
ukraine = ukraine_timemap
QGIS.add_application(ukraine_timemap)



from typing import Dict, List
from argparse import ArgumentParser


def add_default_arguments(
    default_args: Dict[str, Dict],
    parser: ArgumentParser,
):
    for arg, values in default_args.items():
        parser.add_argument(
            f"--{arg}",
            type=values.get("type", str),
            default=values["default"],
            help=values["help"],
        )


def as_list_of_args(default_args: Dict[str, Dict]) -> List[str]:
    return sorted(
        [
            "[--{} {}]".format(
                arg,
                values["help"],
            )
            for arg, values in default_args.items()
        ]
    )



from typing import List

from blueness import module
from blue_objects import host

from blue_geo import NAME
from blue_geo.catalog.EarthSearch.classes import EarthSearchCatalog
from blue_geo.catalog.generic.generic.stac import STACDatacube
from blue_geo.catalog.generic.generic.scope import DatacubeScope
from blue_geo.logger import logger

NAME = module.name(__file__, NAME)


class EarthSearchSentinel2L1CDatacube(STACDatacube):
    catalog = EarthSearchCatalog()

    collection = "sentinel-2-l1c"

    name = "sentinel_2_l1c"

    s3_prefix = "s3://sentinel-s2-l1c/tiles"

    def ingest_filename(
        self,
        filename: str,
        overwrite: bool = False,
        verbose: bool = True,
    ) -> bool:
        if super().ingest_filename(filename, overwrite, verbose):
            return True

        # https://registry.opendata.aws/sentinel-2/
        return host.shell(
            "aws s3 cp --request-payer requester {}/{} {}".format(
                self.s3_prefix,
                filename.replace("_", "/"),
                self.full_filename(filename),
            ),
            log=verbose,
        )

    def list_of_files(
        self,
        scope: DatacubeScope = DatacubeScope("all"),
        verbose: bool = False,
    ) -> List[str]:
        return scope.filter(
            [
                {
                    "filename": asset.href.split(f"{self.s3_prefix}/", 1)[1].replace(
                        "/", "_"
                    ),
                }
                for asset in self.metadata["Item"].assets.values()
                if self.s3_prefix in asset.href
            ],
            needed_for_rgb=lambda filename: filename.endswith("TCI.jp2"),
            is_rgb=lambda filename: filename.endswith("TCI.jp2"),
            verbose=verbose,
        )

    @property
    def raw(self) -> str:
        return "{} | {}".format(
            super().raw,
            self.s3_prefix,
        )



from typing import List, Type

from blue_geo import env
from blue_geo.catalog.generic import (
    GenericCatalog,
    VoidCatalog,
    GenericDatacube,
    VoidDatacube,
)
from blue_geo.catalog.copernicus import CopernicusCatalog, CopernicusSentinel2Datacube
from blue_geo.catalog.EarthSearch import (
    EarthSearchCatalog,
    EarthSearchSentinel2L1CDatacube,
)
from blue_geo.catalog.firms import FirmsCatalog
from blue_geo.catalog.firms.area import FirmsAreaDatacube
from blue_geo.catalog.SkyFox import SkyFoxCatalog
from blue_geo.catalog.SkyFox.Venus import SkyFoxVenusDatacube
from blue_geo.catalog.ukraine_timemap import (
    UkraineTimemapCatalog,
    UkraineTimemapDatacube,
)

list_of_catalog_classes: List[Type[GenericCatalog]] = [
    GenericCatalog,
    CopernicusCatalog,
    EarthSearchCatalog,
    FirmsCatalog,
    SkyFoxCatalog,
    UkraineTimemapCatalog,
]

list_of_catalogs: List[str] = sorted(
    [
        catalog_name
        for catalog_name in [
            catalog_class.name for catalog_class in list_of_catalog_classes
        ]
        if catalog_name not in env.BLUE_GEO_DISABLED_CATALOGS.split(",")
    ]
)

list_of_datacube_classes: List[Type[GenericDatacube]] = [
    GenericDatacube,
    FirmsAreaDatacube,
    UkraineTimemapDatacube,
    CopernicusSentinel2Datacube,
    EarthSearchSentinel2L1CDatacube,
    SkyFoxVenusDatacube,
]



import os
from typing import Any, Tuple, Dict, List
import numpy as np

from blueness import module
from blue_options import string
from blue_objects import file, objects, path
from blue_objects.metadata import post_to_object
from blue_objects.logger.image import log_image_hist
from blue_objects.env import ABCLI_OBJECT_ROOT

from blue_geo import NAME
from blue_geo.catalog.generic.classes import GenericCatalog, VoidCatalog
from blue_geo.catalog.generic.generic.scope import DatacubeScope
from blue_geo.logger import logger

NAME = module.name(__file__, NAME)


class GenericDatacube:
    name = "generic"

    catalog = GenericCatalog()

    datacube_id = ""

    metadata: Any = None

    query_args: Dict[str, Dict] = {
        "arg": {
            "default": "value",
            "help": "<value>",
        },
    }

    QGIS_template = "unknown-template"

    def __init__(self, datacube_id: str = ""):
        self.datacube_id = datacube_id
        if self.datacube_id:
            self.update_metadata()

    @property
    def description(self) -> str:
        return "{}.{}@{}[{}]:".format(
            NAME,
            self.__class__.__name__,
            self.catalog.name,
            self.datacube_id,
        )

    def full_filename(self, filename: str) -> str:
        return objects.path_of(filename, self.datacube_id, create=True)

    def generate(
        self,
        modality: str,
        overwrite: bool = False,
    ) -> str:
        return ""

    def ingest(
        self,
        dryrun: bool = False,
        overwrite: bool = False,
        scope: str = "metadata",
    ) -> Tuple[bool, Any]:
        logger.info(
            "{}.{}.ingest({}): {} @ {}".format(
                NAME,
                self.__class__.__name__,
                ",".join(
                    [
                        item
                        for item in ["dryrun" if dryrun else ""]
                        + ["overwrite" if overwrite else ""]
                        if item
                    ]
                ),
                scope,
                self.datacube_id,
            )
        )

        return True, None

    # returns True if ingest is complete.
    def ingest_filename(
        self,
        filename: str,
        overwrite: bool = False,
        verbose: bool = True,
    ) -> bool:
        item_filename = self.full_filename(filename)

        assert path.create(file.path(item_filename))

        if item_filename.endswith(os.sep):
            return True

        if not overwrite and file.exists(item_filename):
            logger.info(f"✅ {item_filename}")
            return True

        logger.info("ingesting {} ...".format(filename))

        return False

    def list_of_files(
        self,
        scope: DatacubeScope = DatacubeScope("all"),
        verbose: bool = False,
    ) -> List[str]:
        return []

    @staticmethod
    def load_modality(
        filename: str,
        modality: str,
        ignore_error: bool = False,
        log: bool = False,
        verbose: bool = False,
        normalized: bool = True,
    ) -> Tuple[bool, np.ndarray, Dict[str, Any]]:
        success, frame, frame_file_metadata = file.load_geoimage(
            filename,
            ignore_error=ignore_error,
            log=log,
        )
        if not success:
            return success, frame, frame_file_metadata

        frame = np.transpose(frame, (1, 2, 0))

        frame_range = (float(np.min(frame)), float(np.max(frame)))
        logger.info(
            "frame: {} : {}".format(
                string.pretty_shape_of_matrix(frame),
                frame_range,
            )
        )

        if frame.shape[2] == 6:
            frame = frame[:, :, [0, 2, 4]]
        elif frame.shape[2] > 3:
            frame = frame[:, :, :3]

        if log:
            log_image_hist(
                image=frame,
                range=frame_range,
                header=[
                    f"{frame_range[0]:.2f} .. {frame_range[1]:.2f}",
                ],
                footer=(
                    filename.split(f"{ABCLI_OBJECT_ROOT}/", 1)[1].split(os.sep)
                    if ABCLI_OBJECT_ROOT in filename
                    else [file.name_and_extension(filename)]
                ),
                filename=file.add_suffix(
                    file.add_extension(filename, "png"), "histogram"
                ),
            )

        if normalized and frame.dtype == np.uint16:
            frame = frame.astype(np.float32) / 5000 * 255
            frame[frame < 0] = 0
            frame[frame > 255] = 255
            frame = frame.astype(np.uint8)

        return success, frame, frame_file_metadata

    @property
    def raw(self) -> str:
        return "{}.{}: {}".format(
            self.catalog.name,
            self.name,
            self.datacube_id,
        )

    @classmethod
    def parse_datacube_id(cls, datacube_id: str) -> Tuple[
        bool,
        Dict[str, Any],
    ]:
        # datacube-<catalog-name>-<args>
        segments = datacube_id.split("-") + ["", ""]

        return (
            segments[0] == "datacube" and segments[1] == cls.catalog.name,
            {},
        )

    @property
    def path(self) -> str:
        return objects.object_path(self.datacube_id)

    @classmethod
    def query(cls, object_name: str) -> bool:
        logger.info(f"🔎 {cls.__name__}.query -> {object_name}")

        return post_to_object(object_name, "datacube_id", [])

    def raw_datacube_id(
        self,
        datacube_id: str = "",  # to enable upstream modifications
    ) -> str:
        datacube_id = self.datacube_id if not datacube_id else datacube_id

        segments = datacube_id.split("-", 3)

        output = segments[3] if len(segments) >= 4 else ""

        if "-DERIVED-" in output:
            output = output.split("-DERIVED-", 1)[0]

        return output

    def update_metadata(self, verbose: bool = False) -> bool:
        if verbose:
            logger.info(f"{self.description}.update_metadata()")

        return True


class VoidDatacube(GenericDatacube):
    name = "void"
    catalog = VoidCatalog()



from blue_geo.catalog.generic.stac.classes import STACCatalog


class EarthSearchCatalog(STACCatalog):
    name = "EarthSearch"

    url = {
        "api": "https://earth-search.aws.element84.com/v1/",
        "sentinel-2-on-aws-open-data": "https://registry.opendata.aws/sentinel-2/",
        "stac-api-info": "https://stacindex.org/catalogs/earth-search#/",
        "ui": "https://viewer.aws.element84.com/",
    }



from blueness import module
from blue_options import string
from blue_objects.metadata import get_from_object
from blueflow.workflow.generic import Workflow

from blue_geo import NAME
from blue_geo.watch.targets.target import Target
from blue_geo.logger import logger

NAME = module.name(__file__, NAME)


def generate_workflow(
    algo_options: str,
    query_object_name: str,
    job_name: str,
    object_name: str,
    map_options: str,
    reduce_options: str,
) -> bool:
    list_of_datacube_id = get_from_object(
        query_object_name,
        "datacube_id",
    )

    suffix = string.pretty_date(
        include_date=False,
        as_filename=True,
        unique=True,
    )

    logger.info(
        "{}.generate_workflow @ {}: [{} X {} datacube(s)]/{}: -[{} @ {} + {}]-> {}".format(
            NAME,
            algo_options,
            query_object_name,
            len(list_of_datacube_id),
            suffix,
            map_options,
            reduce_options,
            job_name,
            object_name,
        )
    )

    workflow = Workflow(job_name)

    workflow.G.add_node("reduce")
    workflow.G.nodes["reduce"]["command_line"] = " ".join(
        [
            "blueflow_workflow monitor",
            "node=reduce",
            job_name,
            "blue_geo_watch_reduce",
            f",{algo_options},suffix={suffix},{reduce_options}",
            query_object_name,
            object_name,
        ]
    )

    for offset in range(len(list_of_datacube_id)):
        node = f"map-{offset:03d}"

        workflow.G.add_node(node)

        workflow.G.nodes[node]["command_line"] = " ".join(
            [
                "blueflow_workflow monitor",
                f"node={node}",
                job_name,
                "blue_geo_watch_map",
                f",{algo_options},offset={offset:03d},suffix={suffix},{map_options}",
                query_object_name,
            ]
        )

        workflow.G.add_edge("reduce", node)

    return workflow.save()



import os
import re
import yaml
import time
import random
from tqdm import tqdm


if not QGIS_is_live:
    from log import log, log_error, verbose
    from layer import layer
    from project import project
    from seed import seed
    from dependency import list_of_dependencies

    BLUE_GEO_VERSION = "1.1.1"
    ABCLI_OBJECT_ROOT = ""

NAME = "blue_geo.QGIS"


class ABCLI_QGIS(object):
    def __init__(self):
        self.app_list = []

    def add_application(self, app):
        self.app_list += [app]

    def intro(self):
        log(
            "{}-{}: {}".format(
                NAME,
                BLUE_GEO_VERSION,
                ", ".join(
                    [
                        "{} {}".format(
                            app.name,
                            app.icon,
                        )
                        for app in self.app_list
                    ]
                ),
            )
        )
        log(
            'Type in "{}.help()" for help.'.format(
                "|".join(["Q"] + [app.name for app in self.app_list])
            )
        )

    def clear(self):
        # https://gis.stackexchange.com/a/480025/250728
        from qgis.PyQt.QtWidgets import QDockWidget

        consoleWidget = iface.mainWindow().findChild(QDockWidget, "PythonConsole")
        consoleWidget.widget().console.clearButton.trigger()

        self.intro()

        seed("clear")

    def create_video(self, filename="QGIS", object_name=""):
        seed(
            [
                "abcli",
                "create_video",
                f"png,fps=2,filename={filename},gif",
                object_name if object_name else project.name,
            ]
        )

    def export(self, filename="", object_name=""):
        filename = self.file_path(
            filename=filename if filename else "{}.png".format(self.timestamp()),
            object_name=object_name,
        )

        qgis.utils.iface.mapCanvas().saveAsImage(filename)
        log(filename, icon="🖼️")

    def file_path(self, filename, object_name=""):
        return os.path.join(self.object_path(object_name), filename)

    def find_layer(self, layer_name):
        return QgsProject.instance().mapLayersByName(layer_name)

    def get_layer(self, layer_name: str):
        candidate_layers = QgsProject.instance().mapLayersByName(layer_name)

        return candidate_layers[0] if len(candidate_layers) else None

    def help(self, clear=False):
        if clear:
            self.clear()

        log("Q.clear()", "clear Python Console.")
        log("Q.create_video()", "create a video.")
        layer.help()
        if verbose:
            log("Q.export([filename],[object_name])", "export.")
            log("Q.list_of_layers()", "list of layers.")
            log("Q.load(filename,layer_name,template_name)", "load a layer.")
        log('Q.open("|<object-name>|layer|object|project")', "open.")
        project.help()
        if verbose:
            log("Q.refresh()", "refresh.")
            log("Q.reload()", "reload all layers.")
        if verbose:
            log("Q.unload(layer_name)", "unload layer_name.")
        log('Q.upload("|<object-name>|layer|project|qgz")', "upload.")
        log("verbose=True|False", "set verbose state.")

        for app in self.app_list:
            app.help()

    def layer_exists(
        self,
        layer_name,
        do_log: bool = False,
    ):
        for layer_name_ in project.list_of_layers:
            if layer_name_.startswith(layer_name):
                if do_log:
                    log(
                        layer_name,
                        layer_name_ if verbose else "",
                        icon="✅",
                    )
                return True
        return False

    def list_of_layers(self, aux=False):
        output = [
            layer_.name() for layer_ in QgsProject.instance().mapLayers().values()
        ]
        if not aux:
            output = [
                layer_name
                for layer_name in output
                if not layer_name.startswith("Google")
                and not layer_name.startswith("template")
            ]
        log(
            "{} layer(s){}".format(
                len(output),
                ": {}".format(", ".join(output)) if verbose else "",
            ),
            icon="🔎",
        )
        return output

    def load(
        self,
        filename,
        layer_name,
        template_name="",
        refresh=True,
    ):
        if len(self.find_layer(layer_name)) > 0:
            print(f"✅ {layer_name}")
            return True

        if filename.endswith(".geojson"):
            layer_ = QgsVectorLayer(filename, layer_name, "ogr")
        elif filename.endswith(".tif"):
            layer_ = QgsRasterLayer(filename, layer_name)
        else:
            log_error(f"cannot load {filename}.")
            return False

        if not layer_.isValid():
            log_error(f"invalid layer: {filename}.")
            return False

        QgsProject.instance().addMapLayer(layer_)

        if template_name:
            template_layer = self.find_layer(template_name)
            if not len(template_layer):
                log_error(f"template not found: {template_name}.")
                return False

            # https://gis.stackexchange.com/a/357206/210095
            source_style = QgsMapLayerStyle()
            source_style.readFromLayer(template_layer[0])
            source_style.writeToLayer(layer_)
            layer_.triggerRepaint()

        log(
            layer_name,
            template_name,
            icon="🎨",
        )

        if refresh:
            self.refresh()

    def object_path(self, object_name=""):
        output = os.path.join(
            ABCLI_OBJECT_ROOT,
            object_name if object_name else project.name,
        )
        os.makedirs(output, exist_ok=True)
        return output

    def open(self, what="object"):
        self.open_folder(
            layer.path
            if what in "layer"
            else self.object_path() if what == "object" else project.path
        )

    def open_folder(self, path):
        if not path:
            log_error("path not found.")
            return

        log(path)
        os.system(f"open {path}")

    # https://qgis.org/pyqgis/master/core/QgsSettings.html#qgis.core.QgsSettings.allKeys
    # https://docs.qgis.org/3.28/en/docs/pyqgis_developer_cookbook/settings.html
    def recent(self):
        settings = QgsSettings()

        list_of_filenames = [
            settings.value(key)
            for key in settings.allKeys()
            if re.match(r"UI/recentProjects/(\d+)/path", key)
        ]

        output = [
            filename.split(f"{ABCLI_OBJECT_ROOT}/", 1)[1].split("/")[0]
            for filename in list_of_filenames
            if ABCLI_OBJECT_ROOT in filename
        ]

        for filename in list_of_filenames:
            output += list_of_dependencies(filename, ABCLI_OBJECT_ROOT, verbose)

        output = list(set(output))

        filename = os.path.join(ABCLI_OBJECT_ROOT, "QGIS-recent.yaml")
        with open(filename, "w") as file:
            yaml.dump(output, file)
        log(f"-> {filename}")

        return ",".join(output)

    def refresh(self, deep=False):
        log("{}refresh.".format("deep" if deep else ""))
        if deep:
            # https://api.qgis.org/api/classQgsMapCanvas.html
            iface.mapCanvas().redrawAllLayers()
        else:
            iface.mapCanvas().refresh()

    def reload(self):
        # https://gis.stackexchange.com/a/449101/210095
        for layer_ in tqdm(QgsProject.instance().mapLayers().values()):
            layer_.dataProvider().reloadData()

    def remove_layer(
        self,
        layer_name: str,
        refresh: bool = True,
    ):
        for layer_ in QgsProject.instance().mapLayersByName(layer_name):
            QgsProject.instance().removeMapLayer(layer_.id())
            log(layer_name, icon="➖")

        if refresh:
            self.refresh()

    def timestamp(self):
        return time.strftime(
            f"%Y-%m-%d-%H-%M-%S-{random.randrange(100000):05d}",
            time.localtime(time.time()),
        )

    def unload(self, layer_name, refresh=True):
        log(layer_name, icon="🗑️")

        for layer_ in self.find_layer(layer_name):
            QgsProject.instance().removeMapLayer(layer_.id())

        if refresh:
            self.refresh()

    def upload(self, object_name=""):
        seed(
            [
                "abcli_upload",
                f"filename={project.name}.qgz" if object_name == "qgz" else "-",
                (
                    project.name
                    if object_name in ["project", "qgz", project]
                    else (
                        layer.object_name
                        if object_name in ["layer", layer]
                        else (
                            object_name
                            if (isinstance(object_name, str) and object_name)
                            else project.name
                        )
                    )
                ),
            ]
        )


QGIS = ABCLI_QGIS()


def clear():
    QGIS.clear()


def upload(self, object_name=""):
    QGIS.upload(object_name)


Q = QGIS
